
/**********************************************************************
 * Copyright (c) 2005, 2008 IBM Corporation and others.
 * All rights reserved.   This program and the accompanying materials
 * are made available under the terms of the Eclipse Public License v1.0
 * which accompanies this distribution, and is available at
 * http://www.eclipse.org/legal/epl-v10.html
 * $Id: FileManagerLegacy.java,v 1.3 2008-03-20 18:49:59 dmorris Exp $
 * 
 * Contributors: 
 * IBM - Initial API and implementation
 **********************************************************************/

package org.eclipse.hyades.execution.local.file;


import java.io.BufferedInputStream;
import java.io.BufferedOutputStream;
import java.io.File;
import java.io.FileInputStream;
import java.io.FileNotFoundException;
import java.io.FileOutputStream;
import java.io.IOException;
import java.io.InputStream;
import java.io.InterruptedIOException;
import java.io.OutputStream;
import java.net.SocketException;

import org.eclipse.hyades.execution.core.file.IFileLocater;
import org.eclipse.hyades.execution.core.file.IFileLocation;
import org.eclipse.hyades.execution.core.file.IFileManager;
import org.eclipse.hyades.execution.core.impl.ExecutionComponentImpl;
import org.eclipse.hyades.execution.security.ISecureClientParameters;
import org.eclipse.hyades.internal.execution.file.FileServiceConstants;
import org.eclipse.hyades.internal.execution.local.common.ManageFileCommand;
import org.eclipse.hyades.internal.execution.local.common.Message;
import org.eclipse.hyades.internal.execution.local.control.AgentControllerUnavailableException;
import org.eclipse.hyades.internal.execution.local.control.Connection;
import org.eclipse.hyades.internal.execution.local.control.ConnectionImpl;
import org.eclipse.hyades.internal.execution.local.control.Node;
import org.eclipse.hyades.internal.execution.local.control.SecureConnectionImpl;

/**
 * @author Giridhar S
 */
class FileManagerLegacy extends ExecutionComponentImpl implements IFileManager {

	private IFileLocater locator;
	private Connection connection;
	
	FileManagerLegacy(Connection connection) {
		locator=new FileLocaterImpl(connection);
		this.connection=connection;
	}
	
	public void getFile(String localName, String remoteName) throws IOException{
		/* Locate the remote file */
		IFileLocation location=locator.getFile(remoteName);
		
		/* We may have to create the directory if it does not exist */
		String directoryName = null;
		
		File file=new File(localName);
		
		/* We try to get the directory where the file has to be written */
		directoryName = file.getParent();
		File dirFile = new File(directoryName);
		
		/* We have to create the directory only if it does not exist */
		if (!dirFile.exists())
			dirFile.mkdirs();	/* we will have to create parent directories as well */
		
		/* Create the local file.  Should we check if it exists? */
		file.createNewFile();
		
		/* Connect to the remote server and retrieve the file*/
		if(connection instanceof SecureConnectionImpl) {
			SecureResourceConnection resourceConnection=new SecureResourceConnection();
			
			try {
				resourceConnection.setParams(remoteName, ManageFileCommand.GET);
				resourceConnection.setInputStream(null);
				resourceConnection.setOutputStream(new BufferedOutputStream(new FileOutputStream(file)));
				resourceConnection.connect(connection.getNode(), location.getPort());
			}
			catch(AgentControllerUnavailableException e) {
				/* We can ignore this as the same node object is being used
				 * for this connection as the originalconnection to the server.  The
				 * node object has all the proper credential info in its cache
				 */
			}
			catch (FileNotFoundException e1){
				/* This is not expected to happen, since we have created
				 * the file ourselves. But if this does happen for some reason, we should not continue.
				 */
				throw e1;	/* The consumer has to catch this */
			}
		}
		else if(connection instanceof ConnectionImpl) {
			ResourceConnection resourceConnection=new ResourceConnection();
			try {
				resourceConnection.setParams(remoteName, ManageFileCommand.GET);
				resourceConnection.setInputStream(null);
				resourceConnection.setOutputStream(new BufferedOutputStream(new FileOutputStream(file)));
				resourceConnection.connect(connection.getNode(), location.getPort());
			}
			catch(AgentControllerUnavailableException e) {
				/* We can ignore this as the same node object is being used
				 * for this connection as the originalconnection to the server.  The
				 * node object has all the proper credential info in its cache
				 */
			}
			catch (FileNotFoundException e1){
				/* This is not expected to happen, since we have created
				 * the file ourselves. But if this does happen for some reason, we should not continue.
				 */
				throw e1;	/* The consumer has to catch this */
			}
		}
	}
	
	/* (non-Javadoc)
	 * @see org.eclipse.hyades.execution.core.file.IFileManager#deleteFile(java.lang.String)
	 */
	public void deleteFile(String absFileName) throws IOException {
		locator.deleteFile(absFileName);
	}
	
	/* (non-Javadoc)
	 * @see org.eclipse.hyades.execution.core.file.IFileManager#putFile(java.lang.String, java.lang.String)
	 */
	public void putFile(String localName, String remoteName) throws IOException {
		IFileLocation location=locator.putFile(remoteName);
		
		File file=new File(localName);
		
		/* Connect to the remote server and send the file*/
		if(connection instanceof SecureConnectionImpl) {
			SecureResourceConnection resourceConnection=new SecureResourceConnection();
			
			try {
				/* We have to tell the file server the file name and the operation to perform */
				resourceConnection.setParams(remoteName, ManageFileCommand.PUT);
				
				/* The socket's input/output stream is not visible here, since the socket is created
				 * later in the connect method. So if we want to set the input/output stream of the
				 * socket, we set null here, and the socketreader thread checks if a stream is null, and 
				 * sets it to the sockets corresponding stream 
				 */
				try	{
					resourceConnection.setInputStream(new BufferedInputStream(new FileInputStream(file)));
					resourceConnection.setOutputStream(null);
				}
				catch (FileNotFoundException e)	{
					
					/* If the file we are trying to open does not exist, we cannot proceed.
					 * Therefore, we just return.
					 */
					throw e;
					//return;
				}
				
				/* We connect to the file server, and begin the file transfer */
				resourceConnection.connect(connection.getNode(), location.getPort());
			}
			catch(AgentControllerUnavailableException e) {
				/* We can ignore this as the same node object is being used
				 * for this connection as the originalconnection to the server.  The
				 * node object has all the proper credential info in its cache
				 */
			}
		}
		else if(connection instanceof ConnectionImpl) {
			ResourceConnection resourceConnection=new ResourceConnection();
			try {
				/* We have to tell the file server the file name and the operation to perform */
				resourceConnection.setParams(remoteName, ManageFileCommand.PUT);
				
				/* The socket's input/output stream is not visible here, since the socket is created
				 * later in the connect method. So if we want to set the input/output stream of the
				 * socket, we set null here, and the socketreader thread checks if a stream is null, and 
				 * sets it to the sockets corresponding stream 
				 */
				try	{
					resourceConnection.setInputStream(new BufferedInputStream(new FileInputStream(file)));
					resourceConnection.setOutputStream(null);
				}
				catch (FileNotFoundException e)	{
					
					/* If the file we are trying to open does not exist, we cannot proceed.
					 * Therefore, we just return.
					 */
					throw e;
					//return;
				}
				
				/* We connect to the file server, and begin the file transfer */
				resourceConnection.connect(connection.getNode(), location.getPort());
			}
			catch(AgentControllerUnavailableException e) {
				/* We can ignore this as the same node object is being used
				 * for this connection as the originalconnection to the server.  The
				 * node object has all the proper credential info in its cache
				 */
			}
		}
	}
	
	interface IResourceConnection extends Connection{
		public void setParams(String fileName, long operation);
		public void setInputStream(InputStream inStream);
		public void setOutputStream(OutputStream outStream);
		public OutputStream getSocketOutputStream() throws IOException;
		public InputStream getSocketInputStream()throws IOException;
	}
	
	class ResourceConnection extends ConnectionImpl implements IResourceConnection{
		private SocketReader socketReader;
		
		public ResourceConnection()	{
			this.socketReader = new SocketReader(this);
		}
		
		protected void init() {
			/* Set the timeout on the socket reader to be 1 second */
			try {
				_socket.setSoTimeout(1000);
				_socket.setTcpNoDelay(true);
			}
			catch(SocketException e) {
				/* We can ignore this */
			}
			
			/* Currently, the implementation has changed from asynchronous file transfer, to
			 * synchronous file transfer. This is because, clients which have to process the file
			 * after it is transferred completely, are not notified when the file transfer is complete.
			 * When the implementation changes back to asynchronous, we just uncomment this call 
			 */
			
			//new Thread(new SocketReader()).start();
			
			/* For now, we do a synchronous file transfer */
			//(new SocketReader(this)).run();
			socketReader.run();
		}
		
		/* We set the current input stream here */
		public void setInputStream (InputStream in)	{
			socketReader.setInputStream(in);
		}
		
		/* We set the current output stream here */
		public void setOutputStream (OutputStream out)	{
			socketReader.setOutputStream(out);
		}
		
		/* We set the file name and operation, to send to the file server */
		public void setParams(String fileName, long operation)	{
			socketReader.setFileName(fileName);
			socketReader.setOperation(operation);
		}
		
		public OutputStream getSocketOutputStream() throws IOException{
			return _socket.getOutputStream();
		}
		
		public InputStream getSocketInputStream() throws IOException{
			return _socket.getInputStream();
		}
	}
	
	class SecureResourceConnection extends SecureConnectionImpl implements IResourceConnection{
		private SocketReader socketReader;
		
		public SecureResourceConnection()	{
			socketReader = new SocketReader(this);
		}
		
		protected void init() {
			/* Set the timeout on the socket reader to be 1 second */
			try {
				_socket.setSoTimeout(1000);
				_socket.setTcpNoDelay(true);
			}
			catch(SocketException e) {
				/* We can ignore this */
			}
			
			/* Currently, the implementation has changed from asynchronous file transfer, to
			 * synchronous file transfer. This is because, clients which have to process the file
			 * after it is transferred completely, are not notified when the file transfer is complete.
			 * When the implementation changes back to asynchronous, we just uncomment this call 
			 */
			
			//new Thread(new SocketReader()).start();
			
			/* For now, we do a synchronous file transfer */
			//(new SocketReader(this)).run();
			
			socketReader.run();
		}
		
		/* We set the current input stream here */
		public void setInputStream (InputStream in)	{
			socketReader.setInputStream(in);
		}
		
		/* We set the current output stream here */
		public void setOutputStream (OutputStream out)	{
			socketReader.setOutputStream(out);
		}
		
		/* We set the file name and operation, to send to the file server */
		public void setParams(String fileName, long operation)	{
			socketReader.setFileName(fileName);
			socketReader.setOperation(operation);
		}
		
		public OutputStream getSocketOutputStream() throws IOException{
			return _socket.getOutputStream();
		}
		
		public InputStream getSocketInputStream() throws IOException{
			return _socket.getInputStream();
		}
	}
	
	private class SocketReader implements Runnable, FileServiceConstants, org.eclipse.hyades.internal.execution.local.common.Constants {
		private InputStream inStream;
		private OutputStream outStream;
		private String fileName;
		private long operation;

		private boolean _isComplete = false;
		private IResourceConnection connection;
		
		private byte[] buffer=new byte[RA_BUFFER_SIZE];
		
		// David A Scott, davids@us.ibm.com
		// Changing byteCount to a long to accomodate large files
		private long byteCount = 0;
		
		public SocketReader(IResourceConnection connection){
			this.connection = connection;
		}
		
		/* We set the current input stream here */
		public void setInputStream (InputStream in)	{
			this.inStream = in;
		}
		
		/* We set the current output stream here */
		public void setOutputStream (OutputStream out)	{
			this.outStream = out;
		}
		
		/* We set the file name and operation, to send to the file server */
		public void setFileName(String fileName)	{
			this.fileName = fileName;
		}
		
		public void setOperation(long operation){
			this.operation = operation;
		}
		
		// J.Garcia changed to handle Exceptions
		public void run() {
			try {
				/* bugzilla 65922 z/OS triggered change */
				byte[] tempFileName=fileName.getBytes("UTF-8") ;//$NON-NLS-1$
				/* We first send the file name and operation type over the socket to the file server */
				
				OutputStream os = connection.getSocketOutputStream();
				long len = (int)fileName.length();
				int offset = 0;
				
				offset = Message.writeRALongToBuffer(buffer, 0, operation);
				offset = Message.writeRALongToBuffer(buffer, offset, len);
				System.arraycopy((Object)tempFileName, 0, (Object)buffer, offset, tempFileName.length);
				os.write(buffer);
				
				
				/* 
				 * This is where we check to see if we should set the current input/output stream to the
				 * socket's input/output stream. TODO: Check this again.
				 */ 
				if (inStream == null) {
					inStream = connection.getSocketInputStream();
				}
				
				if (outStream == null) {
					outStream = connection.getSocketOutputStream();
				}
				
				/* Run until finished */
				while(!_isComplete) {
					try {
						/* Get the InputStream for this socket so we can read some data */
						int bytesRead=inStream.read(buffer);
						
						if(bytesRead==-1) {
							// if byteCount is zero, then the file was empty and no transfer occurs?
							if (byteCount > 0) {
								// David A Scott, davids@us.ibm.com
								// We have reached EOF on the input buffer file.
								// All of the contents of the file have been written to the socket.
								// Now we should:
								// (a) create an InputStream to the socket to which we are writing
								// (b) issue a read on the input stream
								//     this will block until the socket goes away (i.e., is closed)
								//     this could be because the server is done or because it died
								InputStream in = connection.getSocketInputStream();
								int result = in.read();
								in.close();
								// Anything other than a -1 (EOF) indicates an error was encountered while deploying
                                if (result != -1 || byteCount == 1) {
                                    throw new IOException();
                                }
							}
							break;
						}
						byteCount += bytesRead;
						outStream.write(buffer, 0, bytesRead);
					}
					catch(InterruptedIOException e) {
						/* This happens as the read has timed out.  Just continue as the socket is still valid.
						 * 
						 */
					}
				}
				
			} catch(IOException e) {
				
				throw new RuntimeException(e);
				
			}  finally {
				Node node = connection.getNode();
				
				ISecureClientParameters params = (node != null ? node.getSecurityParameters(): null);
				
				/* The server is now stopping */
				connection.disconnect();
				
				if (node != null) node.setSecurityParameters(params);
				try {
					outStream.close();
					inStream.close();
				}
				catch(IOException e) {
					/* Do we need to handle this? */
				}
			}
		}
	}
}
