/**********************************************************************
 * Copyright (c) 2008, 2011 IBM Corporation
 * All rights reserved. This content is made available under
 * the terms of the Eclipse Public License v1.0
 * which accompanies this distribution, and is available at
 * http://www.eclipse.org/legal/epl-v10.html       
 * $Id: 
 *
 * Contributors:
 * IBM - Initial contribution
 **********************************************************************/ 

package org.eclipse.hyades.loaders.trace;

import java.util.ArrayList;
import java.util.HashMap;
import java.util.Iterator;
import java.util.LinkedList;
import java.util.List;
import java.util.Map;
import java.util.Queue;

public class ThreadDeadlockDetector {
	
	private static final boolean DEBUG = false;
	
	private static final boolean CACHE_ENABLED = true;

	/** When a thread has been determined to be deadlocked, we include it here with hash value 'true' 
	 * - Either a monContendedEnter or monContendedEntered will set the value to false, if already true
	 * */
	private Map<Integer, Boolean> _threadDeadlockCache = new HashMap<Integer, Boolean>();

	/** Whether or not the deadlock of the given thread has already been reported. */
	private Map<Integer, Boolean> _threadDeadlockReported = new HashMap<Integer, Boolean>();
	
	/** The thread that last owned a particular monitor .
	 * - Does not necessarily mean they currently own it
	 * - HOWEVER, if object id is currently being waited on by another thread, then thread id IS necessarily 
	 *   the current parent. */
	private Map<Long /* object id */, Integer /* thread id*/> _lastMonitorOwner = new HashMap<Long, Integer>();
	
	/** Map to the single object that a thread is waiting for (a thread can only wait on one object at a time).
	 * - If it's in this list, someone else must necessarily own it currently */
	private Map<Integer /* thread id */, Long /* object id */> _threadWants = new HashMap<Integer, Long>();
	
	/** Who wants the given monitor. A single object can be wanted by multiple threads */
	private Map<Long /* object id */, List<Integer> /*thread id*/> _monitorWantedBy = new HashMap<Long, List<Integer>>();
	
	/** What objects a thread is currently holding
	 * - This list is necessarily accurate and up-to-date. If a monitor appears on this
	 *   list then it is currently owned by the given thread.
	 * - This list will not contain _all_ monitors that a thread holds, just those that are 
	 *   currently wanted by other threads. It is impossible to get all threads that a monitor 
	 *   holds, but this is not needed for deadlock detection. */
	private Map<Integer /* thread id*/, List<Long> /* thread holds*/> _threadCurrentlyHolds = new HashMap<Integer, List<Long>>(); 
	
	
	public ThreadDeadlockDetector() {
	}
	
	/** If we found a cycle, then add any threads that depend on that cycle to the deadlocked list */
	private List<DeadlockEntry> calculateAllDeadlockedThreads(List<DeadlockEntry> initCycleList) {
		
		/** Whether or not a given thread is locked (map is used for speed) */
		Map<Integer, Boolean> threadLocked = new HashMap<Integer, Boolean>(initCycleList.size()*2);
		
		/** Threads that are in the queue that need to be marked as deadlocked */
		Queue<Integer> threadsToMark = new LinkedList<Integer>();
		
		List<DeadlockEntry> result = new ArrayList<DeadlockEntry>();
		
		// Add the cycle of deadlocks to the list of threads that are deadlocked
		for(DeadlockEntry e : initCycleList) {
			threadLocked.put(e._lockedThread, true);
			
			threadsToMark.add(e._lockedThread);
		}
		
		// While there are still threads that need to be marked as deadlocked..
		while(threadsToMark.size() > 0) {
			Integer i = threadsToMark.remove();
			
			// For each monitor that this thread holds
			List<Long> threadHolds = _threadCurrentlyHolds.get(i);
			if(threadHolds != null) {
				for(Long mon : threadHolds) {
					
					// Each of the threads that want this monitor are now deadlocked as well
					List<Integer> threads = _monitorWantedBy.get(mon);
					for(Integer t : threads) {
						if(t == i) continue; // Skip self
	
						// If the thread isn't already listed as deadlocked...
						Boolean b = threadLocked.get(t);
						if(b == null || b.booleanValue() == false) {
							// ... then list the thread as deadlocked
							
							threadLocked.put(t, true);
							threadsToMark.add(t);
							
						}
					}
					
				}
			}
			
			DeadlockEntry ce = new DeadlockEntry();
			ce._lockedThread = i;			
			ce._lockedObject = _threadWants.get(i);
			ce._objectLockedByWhichThread = _lastMonitorOwner.get(ce._lockedObject);
			
			result.add(ce);
			
		}
		
		return result;
	}
	
	/** If by following the monitor list that a thread wants, we can arrive from the original thread
	 * that we started the recursion with, then we have a cycle. */
	private List<DeadlockEntry> recurse(int originalThread, int currentThread, int depth, Map<Integer, Boolean> threadVisited) {
				
		if(DEBUG) System.out.println("("+depth+") orig:"+originalThread + " / curr:" +currentThread);
		
		boolean dlFound = false;
		if(CACHE_ENABLED) {
			// If we've already determined that a particular thread is deadlocked, then use that value, no need to recalculate
			// (if the value, 'false' is in the map, we instead treat that false as an uncertain, rather than as a 'not deadlocked' value)
			
			Boolean isCurrThreadDeadlocked = _threadDeadlockCache.get((Integer)currentThread);
			if(isCurrThreadDeadlocked != null && isCurrThreadDeadlocked) {
				dlFound = true;
			}
		} 
		
		// If we've analyzed this thread already, this is a cycle (though the originalThread may not be in the cycle itself, may just be a dependent, but it is still deadlocked)
		Boolean threadVis = threadVisited.get((Integer)currentThread);
		if(threadVis != null && threadVis.booleanValue() == true) {
			dlFound = true;
		}
		threadVisited.put(currentThread, true); // mark this thread as visited		

		
		// What does the current thread want? (if anything)
		Long monitorWanted = _threadWants.get((Integer)currentThread);
		
		// Thread wants nothing, therefore no cycle.
		if(monitorWanted == null) {
			return null;
		}
		
		// Thread wants something - monitorWanted
		// - What thread owns it? (if any) 
		Integer threadThatOwnsMonitor = _lastMonitorOwner.get((Long)monitorWanted);
		
		// No thread owns the wanted object, or the object is already owned by the current thread 
		if(threadThatOwnsMonitor == null) return null;
		if(threadThatOwnsMonitor == currentThread) return null;
		
		List<DeadlockEntry> result = new ArrayList<DeadlockEntry>(); 
		
		if(!dlFound) {
			result = recurse(originalThread, threadThatOwnsMonitor, depth+1, threadVisited);
			
			if(result != null && result.size() > 0) { 
				dlFound = true;
			}
		}
		
		if(dlFound) {
			
			DeadlockEntry c = new DeadlockEntry();
			c._lockedObject = monitorWanted;
			c._lockedThread = currentThread;
			c._objectLockedByWhichThread = threadThatOwnsMonitor;
				
			result.add(c);
			
		}
		
		return result;
				
	}
	
	/** Since this method is called when a method ContEnters or ContEntered, this means
	 * the given thread is not deadlocked.
	 * 
	 * Therefore, if this thread was previously deadlocked, then clear the deadlock cache,
	 * because it may now be inaccurate. */
	private void clearDeadlockCacheIfNeeded(int threadIdRef) {
		Boolean result = _threadDeadlockCache.get((Integer)threadIdRef);
		if(result != null && result.booleanValue() == true) {
			_threadDeadlockCache.clear();
			if(DEBUG) {
				System.out.println("Reseting thread deadlock cache");
			}
		}
		
		if(DEBUG) {
			result = _threadDeadlockReported.get((Integer)threadIdRef);
			if(result != null && result.booleanValue() == true) {
				System.out.println(threadIdRef+" has undeadlocked.");
			}
		}
		
	}
	
	public List<DeadlockEntry> handleMonContendedEnter(int threadIdRef, long objIdRef, int objIdRefOwner) {
		
		if(DEBUG) System.out.println("monContendedEnter -  thread:"+threadIdRef + "  objIdRef:"+objIdRef+"  objIdRefOwner:"+objIdRefOwner);
		
		if(threadIdRef < 0 || objIdRef <0 || objIdRefOwner < 0 ) {
			if(DEBUG) System.out.println("Invalid entry received from handleMonContendedEnter; ignored.");
			return null;
		}
		
		if(threadIdRef == 0 || objIdRefOwner == 0) return null;
		
		// Threads that attempt to acquire monitors are not (yet) deadlocked, so update the cache
		
		if(CACHE_ENABLED) {
			clearDeadlockCacheIfNeeded(threadIdRef);
			_threadDeadlockCache.put((Integer)threadIdRef, false);
		}
		_threadDeadlockReported.put((Integer)threadIdRef, false);
		
		_lastMonitorOwner.put(objIdRef, objIdRefOwner);
		_threadWants.put(threadIdRef, objIdRef);
		
		// Update threadThatCurrentlyHolds
		List<Long> objsHeldByThread = _threadCurrentlyHolds.get(objIdRefOwner);
		if(objsHeldByThread == null ) {
			objsHeldByThread = new ArrayList<Long>();
			_threadCurrentlyHolds.put(objIdRefOwner, objsHeldByThread);
		}
		objsHeldByThread.add(objIdRef);

		// Update monitorWantedBy
		List<Integer> threadsThatWantMonitor = _monitorWantedBy.get((Long)objIdRef);
		if(threadsThatWantMonitor == null ) {
			threadsThatWantMonitor = new ArrayList<Integer>();
			_monitorWantedBy.put(objIdRef, threadsThatWantMonitor);
		}
		threadsThatWantMonitor.add(threadIdRef);
		
		// Check graph for cycle
		List<DeadlockEntry> result = recurse(threadIdRef, threadIdRef, 0, new HashMap<Integer, Boolean>());
		
		if(result != null && result.size() > 0) {
			// If we found a cycle, then add any threads that depend on that cycle to the deadlocked list
			result = calculateAllDeadlockedThreads(result);

			// Remove all threads that were previously reported as deadlocked (no need to double-report)
			for(Iterator<DeadlockEntry> i = result.iterator(); i.hasNext(); ) {
				DeadlockEntry ce = i.next();
				
				Boolean b = _threadDeadlockReported.get((Integer)ce._lockedThread);
				if(b != null && b == true) {
					i.remove();
				}
			}
			
			// Mark the threads as deadlocked in the cache
			for(DeadlockEntry ce : result) {
				_threadDeadlockReported.put((Integer)ce._lockedThread, true);
				if(CACHE_ENABLED) {
					_threadDeadlockCache.put((Integer)ce._lockedThread, true);
				}
			}
		}
		
		if(DEBUG) {
			String b = "";

			if(result != null && result.size() > 0) {
				for(DeadlockEntry ce: result) {
					b += ""+ce._lockedThread+", ";
				}
				b = b.substring(0, b.length()-2);
				System.out.println("New deadlocks reported: "+b);
			}
			
		}
				
		return result;
		
	}
	
	public void handleMonContendedEntered(int threadIdRef, long objIdRef) {

		if(DEBUG) System.out.println("monContendedEntered -  thread:"+threadIdRef + "  objIdRef:"+objIdRef);

		if(threadIdRef < 0 || objIdRef <0) {
			if(DEBUG) System.out.println("Invalid entry received from handleMonContendedEntered; ignored.");
			return;
		}
		
		if(threadIdRef == 0) return;
		
		// Threads that acquire monitors are not deadlocked, so update the cache
		if(CACHE_ENABLED) {		
			clearDeadlockCacheIfNeeded(threadIdRef);			
			_threadDeadlockCache.put((Integer)threadIdRef, false);
		}
		_threadDeadlockReported.put((Integer)threadIdRef, false);
		
		// Thread no longer wants objIdRef, cuz it's got it
		Long whatThreadWanted = _threadWants.remove((Integer)threadIdRef);
		if(whatThreadWanted == null || whatThreadWanted != objIdRef)  {
			return;
		}

		// Update threadThatCurrentlyHolds
		Integer lastOwnerOfObj = _lastMonitorOwner.get(objIdRef); // who owned this object before us
		if(lastOwnerOfObj != null) {
			// remove the object from their list of owned
			List<Long> objsHeldByThread = _threadCurrentlyHolds.get(lastOwnerOfObj);
			
			if(objsHeldByThread != null ) {
				objsHeldByThread.remove((Long)objIdRef);
			}
		}
		
		// Monitor is no longer wanted by the thread
		List<Integer> threadsThatWantMonitor = _monitorWantedBy.get((Long)objIdRef);
		if(threadsThatWantMonitor != null) {
			boolean threadInList = threadsThatWantMonitor.remove((Integer)threadIdRef);
			if(!threadInList); 
		}
		
		// Object was last owned by thread
		_lastMonitorOwner.put(objIdRef, threadIdRef);
		
	}
	
	class DeadlockEntry {
		long _lockedObject;
		int _lockedThread;
		int _objectLockedByWhichThread;
	}

}
