/*******************************************************************************
 * Copyright (c) 2011 IBM Corporation and others.
 * All rights reserved. This program and the accompanying materials 
 * are made available under the terms of the Eclipse Public License v1.0
 * which accompanies this distribution, and is available at
 * http://www.eclipse.org/legal/epl-v10.html
 * $Id: 
 * 
 * Contributors:
 *     IBM Corporation - initial API and implementation
 *******************************************************************************/

package org.eclipse.hyades.trace.ui.internal.export;

import java.io.IOException;
import java.io.OutputStream;
import java.math.BigDecimal;

import org.eclipse.hyades.models.hierarchy.TRCAgentProxy;
import org.eclipse.hyades.models.hierarchy.TRCFilter;
import org.eclipse.hyades.models.hierarchy.TRCNode;
import org.eclipse.hyades.models.hierarchy.TRCOption;
import org.eclipse.hyades.models.trace.TRCAggregatedMethodInvocation;
import org.eclipse.hyades.models.trace.TRCClass;
import org.eclipse.hyades.models.trace.TRCFullMethodInvocation;
import org.eclipse.hyades.models.trace.TRCFullTraceObject;
import org.eclipse.hyades.models.trace.TRCGCEvent;
import org.eclipse.hyades.models.trace.TRCMethod;
import org.eclipse.hyades.models.trace.TRCObjAllocSite;
import org.eclipse.hyades.models.trace.TRCProcess;
import org.eclipse.hyades.models.trace.TRCThread;
import org.eclipse.hyades.models.trace.TRCThreadDeadEvent;
import org.eclipse.hyades.models.trace.TRCThreadRunningEvent;

/** In this class, you will see a series of commented fields inside nearly every output method. These are the fields which 
 * are required to be in the binary-outputed trace data, but which do not seem to be used by the workbench itself.
 *
 * For instance, none of this data is provided in the XML versions of these binary tags, nor do I see references to them
 * in the EMF model. Because they do not seem to be used, I have not set them here. 
 **/
public class BinaryElementOutput implements ITraceElementOutput {

	final private static BinaryElementOutput _instance = new BinaryElementOutput();

	private BinaryElementOutput() {
	}
	
	public static BinaryElementOutput getInstance() {
		return _instance;
	}
	
	/** These BigDecimal calcs are all slow for some reason, so do this minimally */
	private long convTimestampSlow(double d, BigDecimal b) {
		b = b.add(new BigDecimal(d));
		b = b.multiply(new BigDecimal(1000000000));
		return b.longValue();
	}
	
	private long convTime(double d) {
		return (long)(d * 1000000000d);
	}

	public void output(OutputStream os, TRCThreadRunningEvent e) throws IOException {
		TRCThread t = e.getThread();

		CThreadStartMessage m = new CThreadStartMessage();
		m._thread_id = t.getId();
		m._timestamp = convTimestampSlow(e.getTime(), t.getProcess().getAbsoluteTime());

		m._thread_name = t.getName();
		m._group_name = t.getGroupName();
		m._parent_group_name = ""; //$NON-NLS-1$

		m.write(os);

//		long _trans_thread_id;
//		long _trans_obj_id_ref;
//		long _obj_id_ref;
//		String _col_value;
//		String _trace_id_ref;

	}

	public void output(OutputStream os, TRCThreadDeadEvent e) throws IOException {
		TRCThread t = e.getThread();
		
		CThreadEndMessage m = new CThreadEndMessage();
		m._thread_id_ref = t.getId();
		m._timestamp = convTimestampSlow(e.getTime(), t.getProcess().getAbsoluteTime()); 
		
		m.write(os);

//		long _trans_thread_id_ref;
//		String _col_value;
//		String _trace_id_ref;

	}

	public void output(OutputStream os, TRCFullMethodInvocation e, double s, BigDecimal bd, boolean isEntry) throws IOException {
		
		if(isEntry) {
			CMethodEntryMessage m = new CMethodEntryMessage();
			m._thread_id_ref = e.getThread().getId();
//			m._timestamp = convTimestamp(e.getEntryTime(), bd);
			m._timestamp = (long)((e.getEntryTime() + s) * 1000000000d);
			
			m._method_id_ref = e.getMethod().getId();
			m._class_id_ref = e.getMethod().getDefiningClass().getId();
			m._ticket = (int)e.getTicket();
			m._stack_depth = e.getStackDepth();
			m.write(os);
			
			// _trans_thread_id_ref;
//			_trans_obj_id_ref;
//			_thread_cpu_time;
//			_seq_counter;
//			_col_value;
//			_trace_id_ref;

		} else {
			CMethodExitMessage m = new CMethodExitMessage();
			m._thread_id_ref = e.getThread().getId();
			m._method_id_ref = e.getMethod().getId();
			m._class_id_ref = e.getMethod().getDefiningClass().getId();
			m._ticket = (int)e.getTicket();
//			m._timestamp = convTimestamp(e.getExitTime(), bd);
			m._timestamp = (long)((e.getExitTime() + s) * 1000000000d);
			m.write(os);
			
//			long _trans_thread_id_ref;
//			long _thread_cpu_time;
//			long _trans_obj_id_ref;
//			long _obj_id_ref;
//			long _trans_class_id_ref;
//			long _seq_counter;
//			String _col_value;
//			String _trace_id_ref;
		}
		
	}

	public void output(OutputStream os, TRCAggregatedMethodInvocation e, boolean isEntry) throws IOException {
		
		if(isEntry) {
			CAGMethodEntryMessage m = new CAGMethodEntryMessage();
			
			m._thread_id_ref = e.getThread().getId();
			m._method_id_ref = e.getMethod().getId();
			m._base_time = convTime(e.getBaseTime());
			m._min_time = convTime(e.getMinTime());
			m._max_time = convTime(e.getMaxTime());
			m._num_calls = e.getCount();
			m.write(os);
			
		} else {
			CAGMethodExitMessage m = new CAGMethodExitMessage();
			m._thread_id_ref = e.getThread().getId();
			m._method_id_ref = e.getMethod().getId();
			
			m.write(os);
		}		
	}

	public void outputTraceEndTags(TRCProcess process, BigDecimal s, OutputStream os) throws IOException {
		
		CRuntimeShutdownMessage m1 = new CRuntimeShutdownMessage();
		m1._timestamp = convTimestampSlow(process.getStopTime(), s);
		m1.write(os);
		
		CTraceEndMessage m2 = new CTraceEndMessage();
		m2._timestamp = convTimestampSlow(process.getStopTime(), s);
		m2.write(os);

		CAgentDestroyMessage m3 = new CAgentDestroyMessage();
		m3._agent_id_ref = ""; //$NON-NLS-1$
		m3._timestamp = convTimestampSlow(process.getStopTime(), s);
		m3.write(os);		
	}

	public void outputTraceStartTags(OutputStream os, TRCProcess process, TRCAgentProxy ap) throws IOException {
		TRCNode node = ap.getProcessProxy().getNode();
		
		AbstractBinaryMessage mg;
		
		mg = new BFormatHeader();
		mg.write(os);
		
		mg = new BEncodingMessage();
		mg.write(os);

		mg = new BFrequencyMessage();
		mg.write(os);

		CNodeMessage m1 = new CNodeMessage();
		m1._node_id = ""; //$NON-NLS-1$
		m1._hostname = node.getName();
		m1._ip = node.getIpAddress();
		m1._timezone = 0;
		m1._timestamp = node.getAbsoluteTime() == null ? convTimestampSlow(0, process.getAbsoluteTime()) : convTimestampSlow(0, node.getAbsoluteTime());
		m1.write(os);
		
		CProcessCreateMessage m2 = new CProcessCreateMessage();
		m2._process_id = ap.getProcessProxy().getRuntimeId();
		m2._pid = ap.getProcessProxy().getPid();
		m2._node_id_ref = ""; //$NON-NLS-1$
		m2._timestamp = node.getAbsoluteTime() == null ? convTimestampSlow(0, process.getAbsoluteTime()) : convTimestampSlow(0, node.getAbsoluteTime());
		m2.write(os);
		
		CAgentCreateMessage m3 = new CAgentCreateMessage();
		m3._agent_id = ap.getRuntimeId();
		m3._version = ""; //$NON-NLS-1$
		m3._process_id_ref = ""; //$NON-NLS-1$
		m3._name = ap.getName();
		m3._type = ap.getType();
		m3._params = ""; //$NON-NLS-1$
		m3._timestamp = node.getAbsoluteTime() == null ? convTimestampSlow(0, process.getAbsoluteTime()) : convTimestampSlow(0, node.getAbsoluteTime());
		m3.write(os);
		
		CTraceStartMessage m4 = new CTraceStartMessage(); 
		m4._trace_id = process.getId();
		m4._agent_id_ref = ""; //$NON-NLS-1$
		m4._timestamp = convTimestampSlow(0, process.getAbsoluteTime());
//		m4._col_value
		m4.write(os);
	}

	public void output(OutputStream os, TRCFilter f) throws IOException {
		
		CFilterMessage m = new CFilterMessage();
		
		if(f.getOrigPattern() != null) m._pattern = f.getOrigPattern();
		if(f.getOrigMode() != null) m._mode = f.getOrigMode();
		if(f.getOrigGenericPattern() != null) m._gen_pattern = f.getOrigGenericPattern();
		if(f.getOrigMethodPattern() != null) m._meth_pattern = f.getOrigMethodPattern();
		if(f.getOrigMethodMode() != null) m._meth_mode = f.getOrigMethodMode();
		if(f.getOrigMethodGenericPattern() != null) m._meth_gen_pattern = f.getOrigMethodGenericPattern();
		
		// m._trace_id_ref
		m.write(os);
		
	}

	public void output(OutputStream os, TRCOption o) throws IOException {
		COptionMessage m = new COptionMessage();
		m._key = o.getKey();
		m._value = o.getValue();
//		m._trace_id_ref
		m.write(os);
		
	}

	public void outputClassDef(TRCClass c, TRCProcess process, OutputStream os) throws IOException {

		if(c.getName() == null || c.getId() < 0) return;
		
		CClassDefMessage m = new CClassDefMessage();
		m._name = c.getOrigFullName();
		
		if(c.getSourceInfo() != null && c.getSourceInfo().getLocation() != null) {
			String loct = c.getSourceInfo().getLocation();
//			loct = loct.replace("<", "&lt;");
//			loct = loct.replace(">", "&gt;");
			m._source_name = loct;
		}

		m._class_id = c.getId();
		m._timestamp = convTimestampSlow(c.getLoadTime(), process.getAbsoluteTime());

		m.write(os);
		
//		long _trans_thread_id_ref;
//		long _thread_id_ref;
//		int _num_interfaces;
//		String _interface_names;
//		long _trans_class_id;
//		String _class_loader;
//		String _superclass;
//		long _trans_obj_id_ref;
//		long _obj_id_ref;
//		String _access;
//		int _num_static_fields;
//		int _num_methods;
//		int _num_instance_fields;
//		String _col_value;
//		String _trace_id_ref;
		
	}

	public void output(OutputStream os, TRCMethod tm) throws IOException {
		CMethodDefMessage m = new CMethodDefMessage();
		m._name = tm.getOrigName();
		m._sig = tm.getOrigSignature();
		m._start_line = tm.getLineNo();
		m._end_line = (tm.getLineCount()+tm.getLineNo());

		m._method_id = tm.getId();

		m._class_id_ref = tm.getDefiningClass().getId();
		
//		String _visibility;
//		byte _is_native;
//		byte _is_abstract;
//		byte _is_static;
//		byte _is_sync;
//		String _exceptions;
//		String _sig_notation;
//		long _trans_class_id_ref;
//		String _col_value;
//		String _trace_id_ref;

		m.write(os);
		
	}

	public void output(OutputStream os, TRCGCEvent e) throws IOException {
		
		double initTime = e.getOwnerThread().getProcess().getAgent().getStartTime();
		
		if (e.getType().equals("start")) { //$NON-NLS-1$
			
			CGCStartMessage m = new CGCStartMessage(); 
			m._timestamp = convTimestampSlow(e.getTime(), new BigDecimal(initTime));
			m.write(os);
			
//			long _trans_thread_id_ref;
//			long _thread_id_ref;
//			String _col_value;
//			String _trace_id_ref;
			
		
		} else if(e.getType().equals("finish")) { //$NON-NLS-1$
			CGCFinishMessage m = new CGCFinishMessage();
			m._timestamp = convTimestampSlow(e.getTime(), new BigDecimal(initTime));
			m.write(os);

//			long _trans_thread_id_ref;
//			long _thread_id_ref;
//			long _total_object_space;
//			long _used_object_space;
//			long _used_objects;
//			String _col_value;
//			String _trace_id_ref;

		} else {
			return;
		}
		
	}

	public void output(OutputStream os, TRCFullTraceObject o, boolean isAlloc) throws IOException {
		double initTime = o.getProcess().getAgent().getStartTime();
		
		if(o.getSize() == 0) return;
		
		if(isAlloc) {
			CObjAllocMessage m = new CObjAllocMessage();
			m._thread_id_ref = o.getThread().getId();
			m._timestamp = (long)((o.getCreateTime() + initTime)*1000000000d);
//			m._timestamp = convTimestamp(o.getCreateTime(), new BigDecimal(initTime));
			m._obj_id_ref = o.getId();
			m._class_id_ref = Math.abs(o.getIsA().getId());
			m._is_array = 0; // This is correct. The isArray field is not ever set in native code. 
			
			m._size = o.getSize();
			
			TRCObjAllocSite as = o.getObjAllocSite();
			if(as  != null) {
				m._method_id = as.getMethod().getId();
				m._line = as.getLineNumber();
			}
			m.write(os);
			
//			long _trans_thread_id_ref;
//			long _trans_obj_id_ref;
//			long _trans_class_id_ref;
//			String _context_data;
//			String _col_value;
//			String _trace_id_ref;			
			
		} else {
			CObjFreeMessage m = new CObjFreeMessage();
//			m._timestamp = convTimestamp(o.getCollectTime(), new BigDecimal(initTime));
			m._timestamp = (long)((o.getCollectTime() + initTime)*1000000000d);
			
			m._obj_id_ref = o.getId();

			m._obj_age = 0; // // age value is not read/used by workbench
			m.write(os);

//			long _trans_thread_id_ref;
//			long _thread_id_ref;
//			long _trans_obj_id_ref;
//			long _seq_counter;
//			long _stack_depth;
//			String _col_value;
//			String _trace_id_ref;

		}
		
	}

	public void outputRuntimeInit(OutputStream os, TRCNode node, TRCProcess process) throws IOException {
		
		long nodeTime = node.getAbsoluteTime() == null ? convTimestampSlow(0, process.getAbsoluteTime()) : convTimestampSlow(0, node.getAbsoluteTime());
		
		CRuntimeInitDoneMessage m = new CRuntimeInitDoneMessage();
		m._thread_id_ref = 1;
		m._timestamp = nodeTime; 
		m.write(os);
		
//		long _trans_thread_id_ref;
//		String _col_value;
//		String _trace_id_ref;
		
	}

	
	
}
