/*****************************************************************************
 * Copyright (c) 2010, 2011 IBM Corporation.
 * All rights reserved. This program and the accompanying materials
 * are made available under the terms of the Eclipse Public License v1.0
 * which accompanies this distribution, and is available at
 * http://www.eclipse.org/legal/epl-v10.html
 *
 * Contributors:
 *    IBM Corporation - Initial API and implementation
 *
 * $Id: HeapObjectData.java,v 1.3 2010-12-14 22:22:26 jcayne Exp $ 
 *****************************************************************************/

package org.eclipse.tptp.martini.analysis;

import java.io.ByteArrayOutputStream;
import java.io.UnsupportedEncodingException;
import java.lang.ref.WeakReference;
import java.lang.reflect.Array;
import java.lang.reflect.Field;
import java.lang.reflect.Modifier;
import java.util.Iterator;
import java.util.Map;
import java.util.WeakHashMap;
import java.util.concurrent.ConcurrentHashMap;

/**
 * Used to store and analyze the Object to obtain instance data information.
 */
public class HeapObjectData {
	/**
	 * Stores the reference to the Object using the TPTP ID (TId) as the key. Uses WeakReference
	 * to prevent holding onto the Object.
	 */
	private static ConcurrentHashMap<Long, WeakReference<Object>> tidInstDataMap;
	
	/**
	 * Prevents adding the map during initialization by initializing the map in the static block.
	 */
	private static boolean initialized;
	
	/**
	 * Tags used when transferring information about the Object.
	 */
	private static final String PRIMITIVE_TAG = "primitiveResult";
	private static final String ARRAY_COMPLEX_TAG = "complexArrayList";
	private static final String ARRAY_PRIMITIVE_TAG = "primitiveArrayList";
	private static final String COMPLEX_TYPE_TAG = "complexType";
	private static final String COMPLEX_MEMBER_TAG = "complexMember";
	private static final String COMPLEX_PRIMITIVE_MEMBER_TAG = "primitiveComplexMember";
	private static final String VERSION_TAG = "version";
	
	// This is a pre-base64 encoded string that says: 
	//
	//   Failed to encode value. Caught UnsupportedEncodingException: UTF-8
	//
	// Technically, all Java implementations are required to support this encoding so
	// its not allow to happen, but if it does, the workbench will decode this into a 
	// readable message.
	private static final String unsupportedEncodingError = "RmFpbGVkIHRvIGVuY29kZSB2YWx1ZS4gQ2F1Z2h0IFVuc3VwcG9ydGVkRW5jb2RpbmdFeGNlcHRpb246IFVURi04";
	
	/**
	 * The agent version.
	 */
	private static final String VERSION = "4.7.3";
	
	/**
	 * Stores the time of the previous run to prevent garbage collection from happening too often.
	 */
	private static long previousRunTime = 0;
	
	/**
	 * Determines if Debug information should be printed.
	 */
	private static final boolean IS_DEBUG_ENABLED = showDebug();
	
	static {
		initialized = false;
		tidInstDataMap = new ConcurrentHashMap<Long, WeakReference<Object>>();
		initialized = true;
	}
	
	/**
	 * Stores an Object with the TPTP ID (TId) in the map.
	 * @param TId The TId of the Object.
	 * @param obj The Object to store in the map.
	 */
	public static void addToInstDataMap(long TId, Object obj)
	{
		// We don't want to record ourself
		if( !initialized || HeapObjectData.class == obj )
			return;
		
		Long tptpId = new Long(TId);
		
		// Synchronization handled by the map.
		tidInstDataMap.put(tptpId, new WeakReference<Object>(obj));
	}
	
	/**
	 * Returns the Object referenced by TId stored in the map.
	 * @param TId The TPTP id of the Object to lookup.
	 * @return The Object with the associated TId.
	 */
	public static Object getObjectFromInstDataMap(long TId)
	{
		// Synchronization handled by the map.
		Long tptpId = new Long(TId);
		if(tidInstDataMap == null)
			return null;

		WeakReference<Object> obj = tidInstDataMap.get(tptpId);
		if(obj == null)
			return null;
		Object objTId = obj.get();
		return objTId;
	}
	
	/**
	 * Returns the TId referenced by the Object.
	 * @param obj The Object to obtain the TId for.
	 * @param create true if a new TId should be created on demand
	 * @return The TId associated with the Object.
	 */
	public static native long getObjectTId(Object obj, boolean create);
	
	/**
	 * Wrapper around {@link HeapObjectData#getObjectTId(Object, boolean)}.
	 *
	 * Provides same API and behaviour where a TId is NOT created if one doesn't
	 * already exist.
	 */
	public static long getObjectTId(Object obj) {
		return getObjectTId( obj, false );
	}

	/**
	 * Analyzes the Object with TId and returns the results.
	 * @param TId The TId associated with the Object to lookup.
	 * @return A String representation of the Object's contents.
	 */
	public static String getObjectValuesFromInstDataMap(long TId)
	{
		// Synchronization handled by the map.
		Long tptpId = new Long(TId);
		Object objTId = null;
		// Skips Objects with a TId of 0
		if(tidInstDataMap == null || TId == 0)
			return new String();
		WeakReference<Object> obj = tidInstDataMap.get(tptpId);
		String results = new String();
		if(obj == null)
			return results;
		objTId = obj.get();
		if(objTId != null)
			results = analyzeObject(objTId);
		return results;
	}
	
	/* Not in use.
	 * Removes the Object and TId entries from the maps.
	 * @param TId The TId of the Object to remove from the map. 
	public static void removeObjectFromInstDataMap(long TId)
	{
		Long tptpId = new Long(TId);
		Object obj = getObjectFromInstDataMap(tptpId);
		if(tidInstDataMap != null)
			tidInstDataMap.remove(tptpId);
	}
	*/
	
	/**
	 * Used to cleanup Objects from the map that have been deallocated during garbage collection.
	 */
	public static void removeUnusedObjectsFromInstDataMap() {
		// Synchronization handled by the map.
		if(tidInstDataMap == null)
			return;
		// Run the cleanup at most once per minute
		long currentRunTime = System.currentTimeMillis();
		if((currentRunTime - previousRunTime) > 60000) {
			previousRunTime = currentRunTime;
			for(Iterator<Map.Entry<Long, WeakReference<Object>>> iter = tidInstDataMap.entrySet().iterator(); iter.hasNext(); ) {
				Map.Entry<Long, WeakReference<Object>> tid = iter.next();
				if(tid.getKey() != null && tid.getValue() != null && tid.getValue().get() == null) {
					iter.remove();
				}
			}
		}
	}
	
	/**
	 * Analyzes the Object to return information about the instance.
	 * @param obj The Object to gather information about.
	 * @return Details about the Object.
	 */
	private static String analyzeObject(Object obj) {
		// Append the version and version tag
		StringBuilder result = new StringBuilder("<"); result.append(VERSION_TAG); result.append(">");	
		result.append(VERSION);	
		result.append("</"); result.append(VERSION_TAG); result.append(">");
		
		if(obj == null)
			return result.toString();
		
		Class<?> objClass = obj.getClass();
				
		// For a primitive object, just return the result
		if(isPrimitiveType(objClass)) {
			result.append("<"); result.append(PRIMITIVE_TAG); result.append(">");
			result.append(obj.getClass().getSimpleName()); result.append(":"); 

			if( isCharType(objClass) )
				result.append( encode( obj.toString() ) );
			else
				result.append(obj.toString());
			result.append("</"); result.append(PRIMITIVE_TAG); result.append(">");
		}
		
		// For an array, return the list of TIds
		else if(objClass.isArray()) {
			int objArraySize = Array.getLength(obj);
			// For arrays with a size, otherwise skip
			if(objArraySize > 0) {

				Class<?> arrayType = objClass.getComponentType();

				// Primitive arrays
				if(isPrimitiveType(arrayType)) {

					// Primitives can be printed
					result.append("<"); result.append(ARRAY_PRIMITIVE_TAG); result.append(">");
					// Add the type of the array
					result.append(arrayType.getSimpleName()); result.append(":");

					// Build up a string with comma separated values
					StringBuilder value = new StringBuilder();
					for(int i = 0; i < objArraySize; i++) {
						Object workingObj = Array.get(obj, i);
						if(workingObj != null) {
							// Append results to the list.
							if(i != 0)
								value.append(",");
							value.append(workingObj.toString());
						}
					}
					if( isCharType(arrayType) )
						result.append( encode( value.toString() ) );
					else
						result.append( value.toString() );
					result.append("</"); result.append(ARRAY_PRIMITIVE_TAG); result.append(">");
				}
				// Complex types have their TIDs returned
				else {
					result.append("<"); result.append(ARRAY_COMPLEX_TAG); result.append(">");
					// Go through the array to get the objects
					for(int i = 0; i < objArraySize; i++) {
						// Append results to the list.
						if(i != 0)
							result.append(",");
						result.append(getObjectTId(Array.get(obj, i), true));
					}
					result.append("</"); result.append(ARRAY_COMPLEX_TAG); result.append(">");
				}
			}
		} else if (obj != null) {

			// It's a complex type
			Class<?> cls = objClass;

			result.append("<"); result.append(COMPLEX_TYPE_TAG); result.append(">");
			while( Object.class != cls ) {

				Field[] fieldz = cls.getDeclaredFields();
				for (int i = 0; i < fieldz.length; i++) {

					Object memberObj = null;
					Class<?> memberType = fieldz[i].getType();

					if( Modifier.isStatic( fieldz[i].getModifiers() ) )
						continue;

					// Get access to all the fields
					fieldz[i].setAccessible(true);
					try {
						// Get the value of the member for the current Object
						memberObj = fieldz[i].get(obj);
					}
					catch (IllegalArgumentException err) {
						if(IS_DEBUG_ENABLED) {
							System.err.println("Failed to obtain the member.");
							err.printStackTrace();
						}
					} catch (IllegalAccessException err) {
						if(IS_DEBUG_ENABLED) {
							System.err.println("Unable to access the member.");
							err.printStackTrace();
						}
					}
					// Check the values of the member Object.
					if(memberObj != null) {
						if(isPrimitiveType(memberType)) {
							result.append("<"); result.append(COMPLEX_PRIMITIVE_MEMBER_TAG); result.append(">");
							result.append(memberType.getSimpleName()); result.append(":"); 
							result.append(fieldz[i].getName()); result.append("=");
							if( isCharType(memberType) )
								result.append( encode( memberObj.toString() ) );
							else
								result.append( memberObj.toString() );
							result.append("</"); result.append(COMPLEX_PRIMITIVE_MEMBER_TAG); result.append(">");
						} else {
							// Add the TId of the field with it's name from the map.
							Long TIdvalue = getObjectTId(memberObj, true);
							
							result.append("<"); result.append(COMPLEX_MEMBER_TAG); result.append(">");
							result.append(memberType.getSimpleName()); result.append(":");
							result.append(fieldz[i].getName()); result.append("="); result.append(TIdvalue); 
							result.append("</"); result.append(COMPLEX_MEMBER_TAG); result.append(">");
						}
					}
				}

				cls = cls.getSuperclass();
			}
			result.append("</"); result.append(COMPLEX_TYPE_TAG); result.append(">");
		}
		else {
			if(IS_DEBUG_ENABLED)
				System.err.println("Unrecongized type passed");
		}
		return result.toString();
	}
	
	/**
	 * Determines if the passed object is of a primitive type.
	 * @param cls The Class to check
	 * @return true if the Class is a primitive, false otherwise.
	 */
	private static boolean isPrimitiveType(Class<?> cls) {
		if(cls == null)
			return false;
		if(cls.isPrimitive())
			return true;

		if(cls == Boolean.class || 
		   cls == Byte.class || 
		   cls == Character.class ||
		   cls == Short.class || 
		   cls == Integer.class || 
		   cls == Long.class || 
		   cls == Float.class || 
		   cls == Double.class || 
		   cls == String.class ) {
			return true;
		}
		return false;
	}
	
	/**
	 * When the Environment Variable HEAP_OBJ_DATA_DEBUG is set to true, this will enable
	 * debug information to be printed.
	 * @return true when HEAP_OBJ_DATA_DEBUG is "true".
	 */
	private static boolean showDebug() {
		String debugEnv = System.getenv("HEAP_OBJ_DATA_DEBUG");
		if(debugEnv != null && debugEnv.equalsIgnoreCase("true"))
			return true;
		return false;
	}

	private static boolean isCharType( Class cls ) {

		if( String.class == cls )
			return true;
		else if( Character.TYPE == cls )
			return true;
		else if( Character.class == cls )
			return true;
		else 
			return false;

	}

	private static String encode( String s ) {

		// We base64 encode Strings and character arrays
		try {
			return encode( s.getBytes("UTF-8") );
		} catch( UnsupportedEncodingException e ) {
			return unsupportedEncodingError;
		}
		
	}

	// Base64 encoder ////////////////////////////////////////////////////////
	
	private static final String mappings = "ABCDEFGHIJKLMNOPQRSTUVWXYZabcdefghijklmnopqrstuvwxyz0123456789+/";//$NON-NLS-1$
	private static final String filler = "=";//$NON-NLS-1$
	
	/**
	 *	Answer a string representing the Base 64 encoded form of the passed
	 *	byte array
	 *
	 *	@return java.lang.String
	 *	@param contents byte[]
	 */
	public static String encode(byte[] contents) {
		
		StringBuffer result = new StringBuffer();
	
		for (int i = 0; i < contents.length; i = i + 3) {
			
			// output character 1
			result.append(mappings.charAt((contents[i] & 0xFC) >> 2));
	
			// output character 2
			int c2 = (contents[i] & 0x03) << 4;
			if (i + 1 >= contents.length) {
				result.append(mappings.charAt(c2));
				result.append(filler);
				result.append(filler);
				return result.toString();
			}
			
			c2 |= ((contents[i + 1] & 0xF0) >> 4);
			result.append(mappings.charAt(c2));
	
			// output character 3
			int c3 = (contents[i + 1] & 0x0F) << 2;
			if (i + 2 >= contents.length) {
				result.append(mappings.charAt(c3));
				result.append(filler);
				return result.toString();
			}
			
			c3 |= ((contents[i + 2] & 0xC0) >> 6);
			result.append(mappings.charAt(c3));
	
			// output character 4
			result.append(mappings.charAt(contents[i + 2] & 0x3F));
		}
		return result.toString();
		
	}
	
	public static String decode( byte[] contents ) throws UnsupportedEncodingException {
		
		ByteArrayOutputStream baos = new ByteArrayOutputStream( 3*contents.length/4 + 4 );
		
		for( int i=0; i<contents.length; i=i+4 ) {
			
			int i1 = mappings.indexOf( contents[ i + 0 ] );
			int i2 = mappings.indexOf( contents[ i + 1 ] );
			
			int c1 = ((i1 & 0x3f) << 2) | ((i2 & 0x3f) >> 4);
			baos.write( c1 );

			if( i + 2 >= contents.length )
				break;
			int i3 = mappings.indexOf( contents[ i + 2 ] );
			if( i3 < 0 )
				break; // padding

			int c2 = ((i2 & 0x0f) << 4) | ((i3 >> 2) & 0x0f);
			baos.write( c2 );

			if( i + 3 >= contents.length )
				break;
			int i4 = mappings.indexOf( contents[ i + 3 ] );
			if( i4 < 0 )
				break; // padding
			
			int c3 = ((i3 & 0x03) << 6) | ((i4 >> 0) & 0x3f);
			baos.write( c3 );
		}
		
		return new String( baos.toByteArray(), "UTF-8" );
		
	}
	
}
