/*******************************************************************************
 * Copyright (c) 2005, 2009 IBM Corporation and others.
 * All rights reserved. This program and the accompanying materials 
 * are made available under the terms of the Eclipse Public License v1.0
 * which accompanies this distribution, and is available at
 * http://www.eclipse.org/legal/epl-v10.html
 * $Id: LaunchConfigurationExtensionsManager.java,v 1.12 2009-05-11 12:50:27 paules Exp $
 * 
 * Contributors:
 *     IBM Corporation - initial API and implementation
 *******************************************************************************/
package org.eclipse.hyades.test.core.internal.launch.extensions;

import java.util.ArrayList;
import java.util.HashMap;
import java.util.Iterator;
import java.util.LinkedList;
import java.util.List;
import java.util.Map;

import org.eclipse.core.runtime.CoreException;
import org.eclipse.core.runtime.IConfigurationElement;
import org.eclipse.core.runtime.IExtensionPoint;
import org.eclipse.core.runtime.Platform;
import org.eclipse.hyades.models.common.testprofile.TPFTest;
import org.eclipse.hyades.models.common.testprofile.TPFTestComponent;
import org.eclipse.hyades.test.core.TestCorePlugin;
import org.eclipse.hyades.test.core.launch.configurations.TestLaunchConfigurationFacade;
import org.eclipse.hyades.test.core.launch.extensions.IDeploymentsProvider;
import org.eclipse.hyades.test.core.launch.extensions.IExecutionHistoryDefaultsProvider;
import org.eclipse.hyades.test.core.launch.extensions.IExecutionHistoryDefaultsProvider2;
import org.eclipse.hyades.test.core.launch.extensions.IExecutionHistoryDefaultsProviderExtended;
import org.eclipse.hyades.test.core.launch.extensions.IExecutionResultsLocationFilter;
import org.eclipse.hyades.test.core.launch.extensions.IRunHandler;
import org.eclipse.hyades.test.core.launch.extensions.ITestLaunchConfigurationValidator;

/**
 * <p>This class is the central point for registering and getting extensions, based
 * on the type and the kind of the Launchable Element.</p>
 * 
 * 
 * @author  Julien Canches
 * @author  Paul Slauenwhite
 * @author  Jeffrey Nevicosi
 * @version May 12, 2009
 * @since   February 1, 2005
 */
public class LaunchConfigurationExtensionsManager {

	private static LaunchConfigurationExtensionsManager instance;

	public static LaunchConfigurationExtensionsManager getInstance() {
		if (instance == null) {
			instance = new LaunchConfigurationExtensionsManager();
		}
		return instance;
	}

	public static final String KIND_TEST = "test"; //$NON-NLS-1$

	public static final String KIND_COMPONENT = "testComponent"; //$NON-NLS-1$

	private LaunchConfigurationExtensionsManager() {
		registerRunHandlers();
		registerDeploymentsProviders();
		registerExecutionHistoryDefaultsProviders();
		registerExecutionHistoryDefaults2Providers();
		registerSupportedTestTypes();
		registerLaunchConfigurationValidators();
		registerExecutionResultsLocationFilters();
	}

	private Map testExtensionMap = new HashMap();

	private Map componentExtensionMap = new HashMap();

	private Map launchConfigurationMap = new HashMap();

	private TestExtensionDescriptor getExtensionDescriptor(String type,
			String kind) {

		if (kind == null || type == null) {
			throw new NullPointerException("Type and kind must be non-null"); //$NON-NLS-1$
		}

		Map map = null;
		if (kind.equals(KIND_TEST)) {
			map = testExtensionMap;
		} else if (kind.equals(KIND_COMPONENT)) {
			map = componentExtensionMap;
		} else {
			throw new IllegalArgumentException("Invalid kind");
		}

		Object o = map.get(type);
		if (o == null) {
			o = new TestExtensionDescriptor();
			map.put(type, o);
		}
		return (TestExtensionDescriptor) o;
	}

	private TestExtensionDescriptor getTestExtensionDescriptor(
			Object testElement) {
		if (testElement != null) {
			if (testElement instanceof TPFTest) {
				TPFTest test = (TPFTest) testElement;
				if (test.getType() != null) {
					return getExtensionDescriptor(test.getType(), KIND_TEST);
				}
			} else if (testElement instanceof TPFTestComponent) {
				TPFTestComponent testComponent = (TPFTestComponent) testElement;
				if (testComponent.getType() != null) {
					return getExtensionDescriptor(testComponent.getType(),
							KIND_COMPONENT);
				}
			}
		}
		return null;
	}

	private class TestExtensionDescriptor {
		public IConfigurationElement runHandlerConfigElem = null;

		private IRunHandler runHandler = null;

		public List supportedLaunchModes = new ArrayList(3);

		public List availableShortcutModes = new ArrayList(3);

		public String launchConfigurationType = null;

		public IConfigurationElement execHistoryDefaultsProviderConfigElem = null;

		private IExecutionHistoryDefaultsProvider execHistoryDefaultsProvider = null;

		public IConfigurationElement deploymentsProviderConfigElem = null;

		private IDeploymentsProvider deploymentsProvider = null;

		public IConfigurationElement launchConfigurationValidatorConfigElem = null;

		private ITestLaunchConfigurationValidator launchConfigurationValidator = null;

		public IRunHandler getRunHandler() {
			if (this.runHandler == null) {
				if (this.runHandlerConfigElem != null) {
					try {
						this.runHandler = (IRunHandler) this.runHandlerConfigElem
								.createExecutableExtension("class"); //$NON-NLS-1$
					} catch (CoreException e) {
						TestCorePlugin.getDefault().logError(e);
						this.runHandlerConfigElem = null;
					}
				}
			}
			return this.runHandler;
		}

		public IExecutionHistoryDefaultsProvider getExecutionHistoryDefaultsProvider() {
			if (this.execHistoryDefaultsProvider == null) {
				if (this.execHistoryDefaultsProviderConfigElem != null) {
					try {
						this.execHistoryDefaultsProvider = (IExecutionHistoryDefaultsProvider) this.execHistoryDefaultsProviderConfigElem
								.createExecutableExtension("class"); //$NON-NLS-1$
					} catch (CoreException e) {
						TestCorePlugin.getDefault().logError(e);
						this.execHistoryDefaultsProviderConfigElem = null;
					}
				}
			}
			return this.execHistoryDefaultsProvider;
		}

		public IDeploymentsProvider getDeploymentsProvider() {
			if (this.deploymentsProvider == null) {
				if (this.deploymentsProviderConfigElem != null) {
					try {
						this.deploymentsProvider = (IDeploymentsProvider) this.deploymentsProviderConfigElem
								.createExecutableExtension("class"); //$NON-NLS-1$
					} catch (CoreException e) {
						TestCorePlugin.getDefault().logError(e);
						this.deploymentsProviderConfigElem = null;
					}
				}
			}
			return this.deploymentsProvider;
		}

		public ITestLaunchConfigurationValidator getLaunchConfigurationValidator() {
			if (this.launchConfigurationValidator == null) {
				if (this.launchConfigurationValidatorConfigElem != null) {
					try {
						this.launchConfigurationValidator = (ITestLaunchConfigurationValidator) this.launchConfigurationValidatorConfigElem
								.createExecutableExtension("class"); //$NON-NLS-1$
					} catch (CoreException e) {
						TestCorePlugin.getDefault().logError(e);
						this.launchConfigurationValidatorConfigElem = null;
					}
				}
			}
			return this.launchConfigurationValidator;
		}

		public boolean isLaunchConfigurationTypeSupported(String lcType) {
			return (this.launchConfigurationType == lcType)
					|| (lcType != null && lcType
							.equals(this.launchConfigurationType));
		}
	}

	private LaunchConfigurationExtensionDescriptor getLaunchConfigurationExtensionDescriptor(
			String launchConfigurationType) {
		
		if (launchConfigurationType == null) {
			throw new NullPointerException("Launch configuration type must be non-null"); //$NON-NLS-1$
		}

		Object o = this.launchConfigurationMap.get(launchConfigurationType);
		if (o == null) {
			o = new LaunchConfigurationExtensionDescriptor();
			this.launchConfigurationMap.put(launchConfigurationType, o);
		}
		return (LaunchConfigurationExtensionDescriptor) o;
	}

	private class LaunchConfigurationExtensionDescriptor {
		public IConfigurationElement execHistoryDefaultsProviderConfigElem = null;

		private IExecutionHistoryDefaultsProvider2 execHistoryDefaultsProvider = null;

		public IConfigurationElement execResultsLocationFilterConfigElem = null;

		public IExecutionResultsLocationFilter execResultsLocationFilter = null;

		public IExecutionHistoryDefaultsProvider2 getExecutionHistoryDefaultsProvider() {
			if (this.execHistoryDefaultsProvider == null) {
				if (this.execHistoryDefaultsProviderConfigElem != null) {
					try {
						this.execHistoryDefaultsProvider = (IExecutionHistoryDefaultsProvider2) this.execHistoryDefaultsProviderConfigElem
								.createExecutableExtension("class"); //$NON-NLS-1$
					} catch (CoreException e) {
						TestCorePlugin.getDefault().logError(e);
						this.execHistoryDefaultsProviderConfigElem = null;
					}
				}
			}
			return this.execHistoryDefaultsProvider;
		}

		public IExecutionResultsLocationFilter getExecutionResultsLocationFilter() {
			if (this.execResultsLocationFilter == null) {
				if (this.execResultsLocationFilterConfigElem != null) {
					try {
						this.execResultsLocationFilter = (IExecutionResultsLocationFilter) this.execResultsLocationFilterConfigElem
								.createExecutableExtension("class"); //$NON-NLS-1$
					} catch (CoreException e) {
						TestCorePlugin.getDefault().logError(e);
						this.execResultsLocationFilterConfigElem = null;
					}
				}
			}
			return this.execResultsLocationFilter;
		}
	}

	private IDeploymentsProvider defaultDeploymentsProvider = new DefaultDeploymentsProvider();

	private IExecutionHistoryDefaultsProvider defaultExecutionHistoryDefaultsProvider = new DefaultExecutionHistoryDefaultsProvider();

	private void registerRunHandlers() {
		IExtensionPoint extPoint = Platform.getExtensionRegistry()
				.getExtensionPoint(
						TestCorePlugin.getPluginId()
								+ ".launchconfigRunHandler"); //$NON-NLS-1$
		if (extPoint != null) {
			IConfigurationElement[] runHandlers = extPoint
					.getConfigurationElements();
			for (int i = 0; i < runHandlers.length; i++) {
				IConfigurationElement[] supportedTypes = runHandlers[i]
						.getChildren();
				for (int j = 0; j < supportedTypes.length; j++) {
					String type = supportedTypes[j].getAttribute("type"); //$NON-NLS-1$
					String kind = supportedTypes[j].getAttribute("kind"); //$NON-NLS-1$
					try {
						TestExtensionDescriptor ed = getExtensionDescriptor(
								type, kind);
						ed.runHandlerConfigElem = runHandlers[i];
					} catch (Exception e) {
						TestCorePlugin
								.getDefault()
								.logError(
										"Extension "
												+ runHandlers[i].getName()
												+ " was ignored. See next messages for details.");
						TestCorePlugin.getDefault().logError(e);
					}
				}
			}
		}
	}

	private void registerExecutionHistoryDefaultsProviders() {
		IExtensionPoint extPoint = Platform.getExtensionRegistry()
				.getExtensionPoint(
						TestCorePlugin.getPluginId()
								+ ".launchconfigExecutionHistoryDefaults"); //$NON-NLS-1$
		if (extPoint != null) {
			IConfigurationElement[] defaultsProviders = extPoint
					.getConfigurationElements();
			for (int i = 0; i < defaultsProviders.length; i++) {
				IConfigurationElement[] supportedStuff = defaultsProviders[i]
						.getChildren();
				for (int j = 0; j < supportedStuff.length; j++) {
					String type = supportedStuff[j].getAttribute("type"); //$NON-NLS-1$
					String kind = supportedStuff[j].getAttribute("kind"); //$NON-NLS-1$
					try {
						TestExtensionDescriptor ed = getExtensionDescriptor(
								type, kind);
						ed.execHistoryDefaultsProviderConfigElem = defaultsProviders[i];
					} catch (Exception e) {
						TestCorePlugin
								.getDefault()
								.logError(
										"Extension "
												+ defaultsProviders[i]
														.getName()
												+ " was ignored. See next messages for details.");
						TestCorePlugin.getDefault().logError(e);
					}
				}
			}
		}
	}

	private void registerExecutionHistoryDefaults2Providers() {
		IExtensionPoint extPoint = Platform.getExtensionRegistry()
				.getExtensionPoint(
						TestCorePlugin.getPluginId()
								+ ".launchconfigExecutionHistoryDefaults2"); //$NON-NLS-1$
		if (extPoint != null) {
			IConfigurationElement[] defaultsProviders = extPoint
					.getConfigurationElements();
			for (int i = 0; i < defaultsProviders.length; i++) {
				IConfigurationElement[] supportedStuff = defaultsProviders[i]
						.getChildren();
				for (int j = 0; j < supportedStuff.length; j++) {
					String type = supportedStuff[j].getAttribute("type");
					try {
						LaunchConfigurationExtensionDescriptor ed = getLaunchConfigurationExtensionDescriptor(type);
						ed.execHistoryDefaultsProviderConfigElem = defaultsProviders[i];
					} catch (Exception e) {
						TestCorePlugin
								.getDefault()
								.logError(
										"Extension "
												+ defaultsProviders[i]
														.getName()
												+ " was ignored. See next messages for details.");
						TestCorePlugin.getDefault().logError(e);
					}
				}
			}
		}
	}

	private void registerDeploymentsProviders() {
		IExtensionPoint extPoint = Platform.getExtensionRegistry()
				.getExtensionPoint(
						TestCorePlugin.getPluginId()
								+ ".launchconfigDeploymentProvider"); //$NON-NLS-1$
		if (extPoint != null) {
			IConfigurationElement[] deploymentProviders = extPoint
					.getConfigurationElements();
			for (int i = 0; i < deploymentProviders.length; i++) {
				IConfigurationElement[] supportedTypes = deploymentProviders[i]
						.getChildren();
				for (int j = 0; j < supportedTypes.length; j++) {
					String type = supportedTypes[j].getAttribute("type"); //$NON-NLS-1$
					String kind = supportedTypes[j].getAttribute("kind"); //$NON-NLS-1$
					try {
						TestExtensionDescriptor ed = getExtensionDescriptor(
								type, kind);
						ed.deploymentsProviderConfigElem = deploymentProviders[i];
					} catch (Exception e) {
						TestCorePlugin
								.getDefault()
								.logError(
										"Extension "
												+ deploymentProviders[i]
														.getName()
												+ " was ignored. See next messages for details.");
						TestCorePlugin.getDefault().logError(e);
					}
				}
			}
		}
	}

	private void registerLaunchConfigurationValidators() {
		IExtensionPoint extPoint = Platform
				.getExtensionRegistry()
				.getExtensionPoint(
						TestCorePlugin.getPluginId() + ".launchconfigValidator"); //$NON-NLS-1$
		if (extPoint != null) {
			IConfigurationElement[] launchConfigurationValidators = extPoint
					.getConfigurationElements();
			for (int i = 0; i < launchConfigurationValidators.length; i++) {
				IConfigurationElement[] supportedTypes = launchConfigurationValidators[i]
						.getChildren();
				for (int j = 0; j < supportedTypes.length; j++) {
					String type = supportedTypes[j].getAttribute("type"); //$NON-NLS-1$
					String kind = supportedTypes[j].getAttribute("kind"); //$NON-NLS-1$
					try {
						TestExtensionDescriptor ed = getExtensionDescriptor(
								type, kind);
						ed.launchConfigurationValidatorConfigElem = launchConfigurationValidators[i];
					} catch (Exception e) {
						TestCorePlugin
								.getDefault()
								.logError(
										"Extension "
												+ launchConfigurationValidators[i]
														.getName()
												+ " was ignored. See next messages for details.");
						TestCorePlugin.getDefault().logError(e);
					}
				}
			}
		}
	}

	private void registerExecutionResultsLocationFilters() {
		IExtensionPoint extPoint = Platform
				.getExtensionRegistry()
				.getExtensionPoint(
						TestCorePlugin.getPluginId()
								+ ".launchconfigExecutionHistoryLocationFilter"); //$NON-NLS-1$
		if (extPoint != null) {
			IConfigurationElement[] execResultsLocationFilters = extPoint
					.getConfigurationElements();
			for (int i = 0; i < execResultsLocationFilters.length; i++) {
				IConfigurationElement[] supportedStuff = execResultsLocationFilters[i]
						.getChildren();
				for (int j = 0; j < supportedStuff.length; j++) {
					String type = supportedStuff[j].getAttribute("type");
					try {
						LaunchConfigurationExtensionDescriptor ed = getLaunchConfigurationExtensionDescriptor(type);
						ed.execResultsLocationFilterConfigElem = execResultsLocationFilters[i];
					} catch (Exception e) {
						TestCorePlugin
								.getDefault()
								.logError(
										"Extension "
												+ execResultsLocationFilters[i]
														.getName()
												+ " was ignored. See next messages for details.");
						TestCorePlugin.getDefault().logError(e);
					}
				}
			}
		}
	}

	private void registerSupportedTestTypes() {
		IExtensionPoint extPoint = Platform.getExtensionRegistry()
				.getExtensionPoint(
						TestCorePlugin.getPluginId()
								+ ".launchconfigLaunchableType"); //$NON-NLS-1$
		if (extPoint != null) {
			IConfigurationElement[] supportedTypes = extPoint
					.getConfigurationElements();
			for (int i = 0; i < supportedTypes.length; i++) {
				String type = supportedTypes[i].getAttribute("type"); //$NON-NLS-1$
				String kind = supportedTypes[i].getAttribute("kind"); //$NON-NLS-1$
				String lcType = supportedTypes[i]
						.getAttribute("launchConfigurationType"); //$NON-NLS-1$
				if (lcType == null) {
					lcType = TestLaunchConfigurationFacade.LAUNCH_CONFIGURATION_TYPE;
				}
				try {
					TestExtensionDescriptor ed = getExtensionDescriptor(type,
							kind);
					ed.launchConfigurationType = lcType;
					IConfigurationElement[] supportedModes = supportedTypes[i]
							.getChildren();
					for (int j = 0; j < supportedModes.length; j++) {
						String mode = supportedModes[j].getAttribute("mode"); //$NON-NLS-1$
						boolean shortcutAvailable = "true".equals(supportedModes[j].getAttribute("shortcutAvailable")); //$NON-NLS-1$ //$NON-NLS-2$
						ed.supportedLaunchModes.add(mode);
						if (shortcutAvailable) {
							ed.availableShortcutModes.add(mode);
						}
					}
				} catch (Exception e) {
					TestCorePlugin
							.getDefault()
							.logError(
									"Extension "
											+ supportedTypes[i].getName()
											+ " was ignored. See next messages for details.");
					TestCorePlugin.getDefault().logError(e);
				}
			}
		}
	}

	
	/**
	 * Returns the registered launch configuration type for the specified
	 * test element.  If test element's type did not explicitly register a
	 * launch configuration type, then the default launch configuration
	 * type is returned.
	 * 
	 * @param testElement
	 * @return
	 */
	public String getLaunchConfigurationType(Object testElement) {
		TestExtensionDescriptor ed = getTestExtensionDescriptor(testElement);
		if (ed != null) {
			return ed.launchConfigurationType;
		}
		else return TestLaunchConfigurationFacade.LAUNCH_CONFIGURATION_TYPE;
	}

	/**
	 * Returns a Run Handler for the specified testElement. This method may
	 * return null if the specified testElement has no Run Handler associated.
	 * In that case, no pre- or post-run handling is expected.
	 * 
	 * @param testElement
	 * @return
	 */
	public IRunHandler getRunHandler(Object testElement) {
		TestExtensionDescriptor ed = getTestExtensionDescriptor(testElement);
		if (ed != null) {
			return ed.getRunHandler();
		}
		return null;
	}

	/**
	 * Return an ExecutionHistoryDefaultsProvider for the specified testElement.
	 * This method never returns null.
	 * 
	 * @param testElement
	 * @return
	 */
	public IExecutionHistoryDefaultsProvider getExecutionHistoryDefaultsProvider(
			Object testElement) {
		IExecutionHistoryDefaultsProvider result = null;
		TestExtensionDescriptor ed = getTestExtensionDescriptor(testElement);
		if (ed != null) {
			result = ed.getExecutionHistoryDefaultsProvider();
		}
		if (result == null) {
			result = defaultExecutionHistoryDefaultsProvider;
		}
		return result;
	}

	/**
	 * Return an ExecutionHistoryDefaultsProvider for the specified testElement.
	 * This method never returns null.
	 * 
	 * @param testElement
	 * @return
	 */
	public IExecutionHistoryDefaultsProviderExtended getExecutionHistoryDefaultsProviderExtended(
			Object testElement) {
		IExecutionHistoryDefaultsProvider result = null;
		TestExtensionDescriptor ed = getTestExtensionDescriptor(testElement);
		if (ed != null) {
			result = ed.getExecutionHistoryDefaultsProvider();
		}
		if (result == null || ! (result instanceof IExecutionHistoryDefaultsProviderExtended) ) {
			result = defaultExecutionHistoryDefaultsProvider;
		}
		return (IExecutionHistoryDefaultsProviderExtended)result;
	}
	
	/**
	 * Return an ExecutionHistoryDefaultsProvider2 for the specified
	 * launchConfiguration. This method may return null.
	 * 
	 * @param launchConfigurationType
	 * @return
	 */
	public IExecutionHistoryDefaultsProvider2 getExecutionHistoryDefaultsProvider2(
			String launchConfigurationType) {
		IExecutionHistoryDefaultsProvider2 result = null;
		LaunchConfigurationExtensionDescriptor ed = getLaunchConfigurationExtensionDescriptor(launchConfigurationType);
		if (ed != null) {
			result = ed.getExecutionHistoryDefaultsProvider();
		}
		return result;
	}

	/**
	 * Return an IExecutionResultsLocationFilter for the specified
	 * launchConfiguration type. This method may return null.
	 * 
	 * @param launchConfigurationType
	 * @return
	 */
	public IExecutionResultsLocationFilter getExecutionResultsLocationFilter(
			String launchConfigurationType) {
		if (launchConfigurationType == null)
			return null;
		IExecutionResultsLocationFilter result = null;
		LaunchConfigurationExtensionDescriptor ed = getLaunchConfigurationExtensionDescriptor(launchConfigurationType);
		if (ed != null) {
			result = ed.getExecutionResultsLocationFilter();
		}
		return result;
	}

	/**
	 * Return a DeploymentsProvider for the specified testElement. This method
	 * never returns null.
	 * 
	 * @param testElement
	 * @return
	 */
	public IDeploymentsProvider getDeploymentsProvider(Object testElement) {
		IDeploymentsProvider result = null;
		TestExtensionDescriptor ed = getTestExtensionDescriptor(testElement);
		if (ed != null) {
			result = ed.getDeploymentsProvider();
		}
		if (result == null) {
			result = defaultDeploymentsProvider;
		}
		return result;
	}

	/**
	 * Return a TestLaunchConfigurationValidator for the specified testElement.
	 * This method may return null.
	 * 
	 * @param testElement
	 * @return
	 */
	public ITestLaunchConfigurationValidator getLaunchConfigurationValidator(
			Object testElement) {
		ITestLaunchConfigurationValidator result = null;
		TestExtensionDescriptor ed = getTestExtensionDescriptor(testElement);
		if (ed != null) {
			result = ed.getLaunchConfigurationValidator();
		}
		return result;
	}

	/**
	 * Returns whether the specified testElement is launchable by a specific
	 * LaunchConfigurationType in the given mode.
	 * 
	 * @param testElement
	 *            The testElement
	 * @param mode
	 *            The run mode used for launching the testElement (usually run,
	 *            debug or profile)
	 * @param launchConfigurationType
	 *            The launchConfigurationType. Use <code>null</code> for the
	 *            Hyades Basic Launch Configuration Type.
	 * @return
	 */
	public boolean isLaunchableElement(Object testElement, String mode, String launchConfigurationType) {
		TestExtensionDescriptor ed = getTestExtensionDescriptor(testElement);
		if (ed != null) {
			return ed
					.isLaunchConfigurationTypeSupported(launchConfigurationType)
					&& ed.supportedLaunchModes.contains(mode);
		}
		return false;
	}
	/**
	 * Returns whether a shortcut should be available for launching the
	 * specified test type in the given mode.  Not intended for launching test components.
	 * 
	 * @param testTypeID
	 *            The ID of the test type to be run (e.g. "org.eclipse.hyades.test.core.whatever")            
	 * @param mode
	 *            The run mode used for launching the testElement (usually run,
	 *            debug or profile)
	 * @return
	 */
	public boolean isShortcutAvailable(String testTypeID, String mode) {
		return isShortcutAvailable(testTypeID, KIND_TEST, mode);
	}

	/**
	 * Returns whether a shortcut should be available for launching the
	 * specified test type in the given mode.  Not intended for launching test components.
	 * 
	 * @param testTypeID
	 *            The ID of the test type to be run (e.g. "org.eclipse.hyades.test.core.whatever")
	 * @param testKind
	 * 		Either KIND_TEST or KIND_COMPONENT                       
	 * @param mode
	 *            The run mode used for launching the testElement (usually run,
	 *            debug or profile)
	 * @return
	 */
	public boolean isShortcutAvailable(String testTypeID, String testKind, String mode) {
		TestExtensionDescriptor ed = getExtensionDescriptor(testTypeID, testKind);
		if (ed != null) {
			return TestLaunchConfigurationFacade.LAUNCH_CONFIGURATION_TYPE.equals(ed.launchConfigurationType) // We support shortcuts
					// for default LC only
					&& ed.availableShortcutModes.contains(mode);
		}
		return false;
	}

	/**
	 * Returns whether a shortcut should be available for launching the
	 * specified testElement in the given mode.
	 * 
	 * @param testElement
	 *            The testElement
	 * @param mode
	 *            The run mode used for launching the testElement (usually run,
	 *            debug or profile)
	 * @return
	 */
	public boolean isShortcutAvailable(Object testElement, String mode) {
		TestExtensionDescriptor ed = getTestExtensionDescriptor(testElement);
		if (ed != null) {
			return TestLaunchConfigurationFacade.LAUNCH_CONFIGURATION_TYPE.equals(ed.launchConfigurationType) // We support shortcuts
					// for default LC only
					&& ed.availableShortcutModes.contains(mode);
		}
		return false;
	}

	/**
	 * @return Returns the defaultExecutionHistoryDefaultsProvider.
	 */
	public IExecutionHistoryDefaultsProvider getDefaultExecutionHistoryDefaultsProvider() {
		return defaultExecutionHistoryDefaultsProvider;
	}

	/**
	 * @return Returns the defaultDeploymentsProvider.
	 */
	public IDeploymentsProvider getDefaultDeploymentsProvider() {
		return defaultDeploymentsProvider;
	}
    
    private List listeners = new LinkedList();

    public void handleLaunchDialogClosed() {
        fireLaunchDialogClosed();
    }

    private void fireLaunchDialogClosed() {
        Iterator it = listeners.iterator();
        while (it.hasNext()) {
            ((LaunchDisposeListener) it.next()).launchDisposed();
        }
    }

    public void addDisposeListener(LaunchDisposeListener validator) {
        listeners.add(validator);
    }
    
    public void removeDisposeListener(LaunchDisposeListener validator) {
        listeners.remove(validator);
    }

}
